/*
   notme-milter - rejects SMTP HELOs falsely claiming to be the mail server

   Copyright (C) 2004 James R. Williams Zavada

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Questions or comments regarding this program should be sent to:

       e-mail: <notme@jrwz.net>
   paper-mail: James R. Williams Zavada
               201 Union Mills Way
               Cary, NC 27519
               United States of America
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sysexits.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include <pwd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libmilter/mfapi.h>

#include "version.h"
#include "parsecfg.h"

#ifndef TRUE
# define TRUE 1
# define FALSE 0
#endif

#define MAXADDRESSES 512

char *Progname;
char *Version = VERSION;
int Logfacility = LOG_MAIL;
char Config_file[MAXPATHLEN + 1] = "/etc/mail/notme-milter.conf";
char Interface[MAXPATHLEN + 1] = "";
char Reject_rcode[4] = "550";
char Reject_xcode[6] = "5.7.1";
char Reject_message[MAXSTRING + 1] = "Message Rejected - Forged HELO Host Address!";
char Reject_hook[MAXSTRING + 1] = "";

void print_usage(void);
int initialisation(int argc, char *argv[], char *args, char *default_config_file);
int parse_logfacility(char *value);
int parse_interface(char *value);
int parse_user(char *value);
int parse_rejectmsg(char *value);
int parse_rejecthook(char *value);
int validate_ip_address(char *ip_address);

sfsistat nmm_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr);
sfsistat nmm_helo(SMFICTX *ctx, char *helohost);
sfsistat nmm_abort(SMFICTX *ctx);
sfsistat nmm_close(SMFICTX *ctx);


struct options config_options[] =
{
  {"logfacility", parse_logfacility},
  {"interface", parse_interface},
  {"user", parse_user},
  {"rejectmsg", parse_rejectmsg},
  {"rejecthook", parse_rejecthook},
  {NULL, NULL},
};

struct smfiDesc nmmilter =
{
  "Notme-Milter", /* filter name */
  SMFI_VERSION,   /* version code -- do not change */
  0,              /* flags */
  nmm_connect,    /* connection info filter */
  nmm_helo,       /* SMTP HELO command filter */
  NULL,           /* envelope sender filter */
  NULL,           /* envelope recipient filter */
  NULL,           /* header filter */
  NULL,           /* end of header */
  NULL,           /* body block filter */
  NULL,           /* end of message */
  nmm_abort,      /* message aborted */
  nmm_close       /* connection cleanup */
};

struct nmfiPriv
{
  char *clientaddr;
};


int main(int argc, char *argv[])
{
  int status = EX_OK;
  pid_t pid = 0;

  status = initialisation(argc, argv, "h?vc:", Config_file);

  if ( EX_OK == status )
  {
    /* set up milter then do the daemon fork */

    (void)smfi_setconn(Interface);

    if ( MI_FAILURE == smfi_register(nmmilter) )
    {
      fprintf(stderr, "Error: smfi_register(nmmilter) failed\n");

      status = EX_UNAVAILABLE;
    }
    else
    {
      if ( MI_FAILURE == smfi_opensocket(TRUE) )
      {
        fprintf(stderr, "Error: cannot open interface %s\n", Interface);

         status = EX_UNAVAILABLE;
      }
      else
      {
        openlog("notme-milter", LOG_PID|LOG_CONS, Logfacility);

        if ( -1 == ( pid = fork() ) ) /* Cannot fork the daemon! */
        {
          fprintf(stderr, "Error: cannot do daemon fork: %s\n", strerror(errno));

          status = EX_UNAVAILABLE;
        }
        else if ( 0 == pid ) /* I am the spawned daemon child */
        {
          status = smfi_main();

          /* When smfi_main() returns, program is done, log it */
          syslog(LOG_INFO, "Notme-milter version %s exiting, status: %d", Version, status);

          if ( MI_SUCCESS == status )
          {
            status = EX_OK;
          }
        }
        else /* I am the parent */
        {
          syslog(LOG_INFO, "Notme-milter version %s starting with config from %s", Version,
                 Config_file);

          status = EX_OK;
        }  /* End if/elsif/else fork() succeeded (as parent) */
      }  /* End if/else smfi_opensocket() succeeded */
    }  /* End if/else smfi_register() succeeded */
  }  /* End if/else initialisation() succeeded */

  return(status);
}

void print_usage(void)
{
  printf("\nUsage: %s [-h] [-v] [-c /path/to/config_file]\n\n", Progname);
  puts("Where -h displays this help message and exits, -v displays");
  puts("the program version and exits, and -c /path/to/config_file");
  puts("runs the program with a config file other than the default");
  puts("at /etc/mail/notme-milter.conf\n");
}

int initialisation(int argc, char *argv[], char *args, char *default_config_file)
{
  int optc;
  int status = EX_OK;
  char *tmpptr;
  char *config_file = default_config_file;
  int help_flag = 0;
  int version_flag = 0;

  /* First, we set up our Global program name */
  Progname = argv[0];

  /* Then we set up a temp pointer to the last slash in the program name */
  tmpptr = (char *)strrchr(Progname, '/');

  /* If there's no slash, temp pointer points to program name */
  if ( NULL == tmpptr )
  {
    tmpptr = Progname;
  }
  else /* We update temp pointer to character after slash */
  {
    tmpptr++;
  }

  /* And point our program name to the temp pointer (program's basename) */
  Progname = tmpptr;

  /* Flip through any and all command-line args */
  while ( -1 != (optc = getopt(argc, argv, args)) )
  {
    switch( optc )
    {
      case 'h': /* Help */
        print_usage();

        help_flag = 1;

        status = EX_USAGE;
        break;

      case 'c': /* Config file */
        config_file = strdup(optarg);

        if ( NULL == config_file )
        {
          fprintf(stderr, "\nError: unable to allocate memory for config filename\n\n");

          status = EX_USAGE;
        }
        break;

      case 'v': /* User asked for version */
        printf("\nNotme-milter version %s\n\n", Version);

        version_flag = 1;

        status = EX_USAGE;
        break;

      case ':': /* User Error */
        status = EX_USAGE;
        break;

      case '?': /* Unknown Option */
        status = EX_USAGE;
        break;
    }
  } 

  argv += optind;
  argc -= optind;

  /* We should have no remaining args now */
  if ( 0 != argc )
  {
    fprintf(stderr, "\nError: too many command-line arguments\n\n");

    status = EX_USAGE;
  }

  if ( !help_flag &&
       !version_flag )
  {
    strncpy(Config_file, config_file, MAXPATHLEN);

    Config_file[MAXPATHLEN] = '\0';

    /* Now we parse our config file */
    if ( EX_OK != parse_config_file(config_file) )
    {
      status = EX_NOINPUT;
    }
  }

  return(status);
}

int parse_logfacility(char *value)
{
  int status = EX_OK;

  value = strlower(value);

  if ( 0 == strcmp(value, "mail") )
  {
    Logfacility = LOG_MAIL;
  }
  else if ( 0 == strcmp(value, "daemon") )
  {
    Logfacility = LOG_DAEMON;
  }
  else if ( 0 == strcmp(value, "user") )
  {
    Logfacility = LOG_USER;
  }
  else if ( 0 == strcmp(value, "local0") )
  {
    Logfacility = LOG_LOCAL0;
  }
  else if ( 0 == strcmp(value, "local1") )
  {
    Logfacility = LOG_LOCAL1;
  }
  else if ( 0 == strcmp(value, "local2") )
  {
    Logfacility = LOG_LOCAL2;
  }
  else if ( 0 == strcmp(value, "local3") )
  {
    Logfacility = LOG_LOCAL3;
  }
  else if ( 0 == strcmp(value, "local4") )
  {
    Logfacility = LOG_LOCAL4;
  }
  else if ( 0 == strcmp(value, "local5") )
  {
    Logfacility = LOG_LOCAL5;
  }
  else if ( 0 == strcmp(value, "local6") )
  {
    Logfacility = LOG_LOCAL6;
  }
  else if ( 0 == strcmp(value, "local7") )
  {
    Logfacility = LOG_LOCAL7;
  }
  else
  {
    fprintf(stderr, "Error: invalid syslog facility (must be mail, daemon, user or local0-7)\n");

    status = EX_CONFIG;
  }

  return(status);
}

int parse_interface(char *value)
{
  int status = EX_OK;

  strncpy(Interface, value, MAXPATHLEN);

  return(status);
}

int parse_user(char *value)
{
  int status = EX_OK;

  if ( 0 != getuid() )
  {
    fprintf(stderr, "Error: only root can set the user\n");

    status = EX_USAGE;
  }
  else
  {
    struct passwd *pw = NULL;

    if ( NULL == (pw = getpwnam(value)) )
    {
      fprintf(stderr, "Error: cannot get user info for %s: %s", value, strerror(errno));

      status = EX_NOUSER;
    }
    else
    {
      if ( ( 0 != setuid(pw->pw_uid) ) ||
           ( 0 != seteuid(pw->pw_uid) ) )
      {
        fprintf(stderr, "Error: cannot set uid to %d: %s", pw->pw_uid, strerror(errno));

        status = EX_NOUSER;
      }
    }
  }

  return(status);
}

int parse_rejectmsg(char *value)
{
  int status = EX_OK;
  char rcode_buffer[MAXSTRING + 1] = "";
  char xcode_buffer[MAXSTRING + 1] = "";
  char *strptr = NULL;
  char *placeholder = NULL;

  /* Our first segment is delimited by a space */
  strptr = strtok_r(value, " ", &placeholder);

  if ( NULL == strptr )
  {
    status = EX_USAGE;
  }
  else
  {
    /* save a copy of our potential reject code */
    strcpy(rcode_buffer, strptr);

    /* our next field is the extended reject code, delimited by a space */
    strptr = strtok_r(NULL, " ", &placeholder);

    if ( NULL == strptr )
    {
      status = EX_USAGE;
    }
    else
    {
      /* save a copy of our potential extended reject code */
      strcpy(xcode_buffer, strptr);

      /* And our last field is just the rest of the string */
      strptr = strtok_r(NULL, "\0", &placeholder);

      if ( NULL == strptr )
      {
        status = EX_USAGE;
      }
      else
      {
        strcpy(Reject_message, strptr);
      }
    }
  }

  /* Now let's make sure the reject code is a 3-digit number */
  if ( ( 3 != strlen(rcode_buffer) ) ||
         ( ! isdigit(rcode_buffer[0]) ) ||
         ( ! isdigit(rcode_buffer[1]) ) ||
         ( ! isdigit(rcode_buffer[2]) ) )
  {
    status = EX_USAGE;
  }
  else
  {
    strcpy(Reject_rcode, rcode_buffer);
  }

  /* And our extended reject code should also be 3 digits,
     but separated by periods */
  if ( ( 5 != strlen(xcode_buffer) ) ||
         ( ! isdigit(xcode_buffer[0]) ) ||
         ( ! ('.' == xcode_buffer[1]) ) ||
         ( ! isdigit(xcode_buffer[2]) ) ||
         ( ! ('.' == xcode_buffer[3]) ) ||
         ( ! isdigit(xcode_buffer[4]) ) )
  {
    status = EX_USAGE;
  }
  else
  {
    strcpy(Reject_xcode, xcode_buffer);
  }

  return(status);
}

int parse_rejecthook(char *value)
{
  int status = EX_OK;

  /* In order for the reject hook to work, we must have access: */
  if ( 0 != access(value, R_OK | X_OK) )
  {
    fprintf(stderr, "Error: invalid permissions (need read and execute)\n");

    status = EX_NOPERM;
  }
  else
  {
    strncpy(Reject_hook, value, MAXSTRING); 

    Reject_hook[MAXSTRING] = '\0';
  }

  return(status);
}

int validate_ip_address(char *ip_address)
{
  int status = 0;
  int quad_count = 0;
  int quad_value = 0;
  char *ip_address_copy = NULL;
  char *strptr = NULL;
  char *placeholder = NULL;

  ip_address_copy = strdup(ip_address);

  if ( NULL == ip_address_copy )
  {
    status = -1;
  }
  else
  {
    /* Check the ip address, quad by quad */
    strptr = strtok_r(ip_address_copy, ".", &placeholder);

    while ( NULL != strptr )
    {
      /* Keep track of number of quads. */
      quad_count++;

      /* If we've more than 4, it can't be a valid IP address! */
      if ( quad_count > 4 )
      {
        status = 1;

        break;
      }

      /* Convert the quad to a value so we can verify it */
      quad_value = atoi(strptr);

      /* Is it a real IP quad? */
      if ( ( quad_value < 0 ) || ( quad_value > 255 ) )
      {
        status = 1;

        break;
      }

      /* Make sure a zero integer is a zero symbol in our string! */
      if ( ( 0 == quad_value ) &&
           ( 0 != strncmp(strptr, "0", 1) ) )
      {
        status = 1;

        break;
      }

      /* Grab the next quad. */
      strptr = strtok_r(NULL, ".", &placeholder);
    }

    if ( quad_count < 4 )
    {
      status = 1;
    }
  }
  
  return(status);
}

sfsistat nmm_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
  struct nmfiPriv *priv;
  struct sockaddr_in *addr_in = (struct sockaddr_in *)hostaddr;

  /* allocate some private memory */
  priv = malloc(sizeof(*priv));

  memset(priv, '\0', sizeof(*priv));

  priv->clientaddr = strdup(inet_ntoa(addr_in->sin_addr));

  if ( NULL == priv->clientaddr )
  {
    syslog(LOG_ERR, "error: cannot allocate memory for clientaddr");

    free(priv);

    /* can't accept this message right now */
    return(SMFIS_TEMPFAIL);
  }

  syslog(LOG_INFO, "connection from: %s [%s]", hostname, priv->clientaddr);

  /* save the private data */
  smfi_setpriv(ctx, priv);

  return(SMFIS_CONTINUE);
}

sfsistat nmm_helo(SMFICTX *ctx, char *helohost)
{
  struct nmfiPriv *priv = smfi_getpriv(ctx);
  int status = SMFIS_CONTINUE;
  pid_t pid;
  int helo_ipstr_len;
  int client_ipstr_len;

  /* Did the connecting client HELO with an IP address? */
  if ( 0 == validate_ip_address(helohost) )
  {
    /* Yep, it did.  This merits further checking, and to do so we need
       to gather the length of the two IP address strings. */
    helo_ipstr_len = strlen(helohost);
    client_ipstr_len = strlen(priv->clientaddr);

    if (
         /* If the two strings are NOT the same length, OR... */
         ( helo_ipstr_len != client_ipstr_len ) ||
         (
           /* If they are the same length AND the two strings don't match... */
           ( helo_ipstr_len == client_ipstr_len ) &&
           ( 0 != 
	       strncmp(helohost, priv->clientaddr, strlen(priv->clientaddr)) )
         )
       )
    {
      /* Then the connecting client is lying about it's IP address!!! */

      /* if we couldn't set the reply, we have milter return a temp failure */
      if ( MI_FAILURE == 
             smfi_setreply(ctx, Reject_rcode, Reject_xcode, Reject_message) )
      {
        status = SMFIS_TEMPFAIL;
      }
      else /* We set the reply, have milter return a reject & log the
              rejection */
      {
        status = SMFIS_REJECT;

        syslog(LOG_INFO, "connection from %s claiming to be %s REJECTED", 
               priv->clientaddr, helohost);
      }

      /* If there is a configured reject hook program, try and execute it. */
      if ( 0 != strlen(Reject_hook) )
      {
        if ( -1 == ( pid = fork() ) ) /* Fork failed */
        {
          syslog(LOG_ERR, "fork() failed, cannot exec Reject_hook: %s",
                 Reject_hook);
        }
        else if ( 0 == pid )          /* Child */
        {
          syslog(LOG_DEBUG, "running execl(%s, %s, %s, %s, NULL)", Reject_hook,
                 Reject_hook, priv->clientaddr, helohost);

          if (
               -1 == execl(Reject_hook, Reject_hook, priv->clientaddr,
                           helohost, NULL)
             )
          {
            syslog(LOG_ERR, "execl() failed, cannot exec Reject_hook: %s",
                   Reject_hook);
          }
        }
        else                          /* Parent */
        {
          syslog(LOG_INFO, "successfully forked to exec Reject_hook: %s",
                 Reject_hook);
        }
      } /* End: if there's a configured reject hook... */
    } /* End: if the connecting client is lying about it's IP address... */
  } /* End: if the connecting client did a HELO with an IP address... */

  /* Client didn't HELO with an IP address, so have milter return OK &
     log acceptance */
  if ( SMFIS_CONTINUE == status)
  {
    syslog(LOG_INFO, "connection from %s claiming to be %s ACCEPTED", 
           priv->clientaddr, helohost);
  }

  return(status);
}

sfsistat nmm_abort(SMFICTX *ctx)
{
  struct nmfiPriv *priv = smfi_getpriv(ctx);

  syslog(LOG_INFO, "connection from %s aborted", priv->clientaddr);

  return(SMFIS_CONTINUE);
}

sfsistat nmm_close(SMFICTX *ctx)
{
  struct nmfiPriv *priv = smfi_getpriv(ctx);

  syslog(LOG_INFO, "connection from %s closed", priv->clientaddr);

  if ( NULL != priv )
  {
    if ( NULL != priv->clientaddr )
    {
      free(priv->clientaddr);
    }

    free(priv);
  
    smfi_setpriv(ctx, NULL);
  }

  return(SMFIS_CONTINUE);
}

